/*
 * Copyright 2021-2022 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "common.h"
#include "pef.h"
#include "xstatus.h"
#include "xparameters.h"
#ifdef DO_TRACE
#ifndef DO_TRACE_CORESIGHT
#include "xuartpsv.h"
#endif
#endif
#include "xqspipsu.h"
#if defined(r5_0) || defined(r5_1)
#include "xil_mpu.h"
#endif
#include "xbir_qspi_hw.h"
#include "pef_hw.h"
#include "xbir_qspi.h"

MessageArea __attribute__((section (".MessageArea"))) g_tMessageArea;
AuxArea __attribute__((section (".AuxArea0"))) g_tAuxArea0;
AuxArea __attribute__((section (".AuxArea1"))) g_tAuxArea1;

/* ------------------------------------------------------------------ */

static void agentDataBarrier(ProgEnvFuncs* pPEF,CACHE_REGION eCacheRegion)
{
    /*
     * NOTE: Flush does invalidate as well
     */
    switch(eCacheRegion) {
    default: /* FallThrough */
    case CACHE_REGION_COMM_AREA:
        Xil_DCacheFlushRange((INTPTR)&pPEF->pMA->tCommandArea,sizeof(CommandArea));
        break;
    case CACHE_REGION_RESP_AREA:
        Xil_DCacheFlushRange((INTPTR)&pPEF->pMA->tResponseArea,sizeof(ResponseArea));
        break;
    case CACHE_REGION_AUX_AREA:
        Xil_DCacheFlushRange((INTPTR)pPEF->pAA->pbData,MAX_PEB_SIZE_BYTES);
        break;
    case CACHE_REGION_PRIV_AREA:
        Xil_DCacheFlushRange((INTPTR)pPEF->pPA->pbData,MAX_PEB_SIZE_BYTES);
        break;
    }
}

/* ------------------------------------------------------------------ */

static uint32_t agentVerifyFlash(
    ProgEnvFuncs* pPEF,
    volatile uint32_t* pqErrBSP
    )
{
    LoopState* pLS;
	XQspiPsu_Config *pQspiPsuConfig;
    PEFContext* pContext;
    uint32_t qPrescaler;

    TRACE();
    *pqErrBSP = 0;
    pLS = &pPEF->tLS;
    pContext = pPEF->pContext;

	/*
	 * Initialize the QSPIPSU driver so that it's ready to use
	 */
	pQspiPsuConfig = XQspiPsu_LookupConfig(XPAR_XQSPIPSU_0_DEVICE_ID);
    if(!pQspiPsuConfig) {
        TRACE();
        *pqErrBSP = ~0;
        return ERR_BSP;
    }
    TRACE();

    /* Override config based on real board */

        /**< Single, Stacked and Parallel mode */
    switch((CONNECTION_MODE)pLS->tVerifyFlashArgsIn.qConnectionMode) {
    default:    /* FallThrough */
    case CONNECTION_MODE_LOWER:
    TRACE();
        pQspiPsuConfig->ConnectionMode = XQSPIPSU_CONNECTION_MODE_SINGLE;
        break;

    case CONNECTION_MODE_DUAL_PARALLEL:
    TRACE();
        pQspiPsuConfig->ConnectionMode = XQSPIPSU_CONNECTION_MODE_PARALLEL;
        break;
    }

        /**< Bus width available on board */
    switch(pLS->tVerifyFlashArgsIn.qBusWidth) {
    default:
    case 1:
    TRACE();
        pQspiPsuConfig->BusWidth = XBIR_QSPI_BUSWIDTH_ONE;
        break;

    case 2:
    TRACE();
        pQspiPsuConfig->BusWidth = XBIR_QSPI_BUSWIDTH_TWO;
        break;

    case 4:
    TRACE();
        pQspiPsuConfig->BusWidth = XBIR_QSPI_BUSWIDTH_FOUR;
        break;
    }

        /**< Input clock frequency */
    pQspiPsuConfig->InputClockHz = pLS->tVerifyFlashArgsIn.qQspiPeripheralInputSpeedHz;	

        /**< Describes whether Cache Coherent or not */
    pQspiPsuConfig->IsCacheCoherent = pLS->tVerifyFlashArgsIn.qCacheCoherentDMA;

    switch(pLS->tVerifyFlashArgsIn.qPrescaler) {
    default:    /* FallThrough */
    case 256:
    TRACE();
		qPrescaler = XQSPIPSU_CLK_PRESCALE_256;
        break;

    case 128:
    TRACE();
		qPrescaler = XQSPIPSU_CLK_PRESCALE_128;
        break;

    case 64:
    TRACE();
		qPrescaler = XQSPIPSU_CLK_PRESCALE_64;
        break;

    case 32:
    TRACE();
		qPrescaler = XQSPIPSU_CLK_PRESCALE_32;
        break;

    case 16:
    TRACE();
		qPrescaler = XQSPIPSU_CLK_PRESCALE_16;
        break;

    case 8:
    TRACE();
		qPrescaler = XQSPIPSU_CLK_PRESCALE_8;
        break;

    case 4:
    TRACE();
		qPrescaler = XQSPIPSU_CLK_PRESCALE_4;
        break;

    case 2:
    TRACE();
		qPrescaler = XQSPIPSU_CLK_PRESCALE_2;
        break;
    }

    TRACE();
    *pqErrBSP = FlashInit(&pContext->tXbirContext,pQspiPsuConfig,qPrescaler);
    if((u32)XST_SUCCESS != *pqErrBSP) {
        TRACE();
        return ERR_BSP;
    }

    TRACE();
    if(0
            || (
                    pLS->tVerifyFlashArgsIn.qPEBSize
                    !=
                    pContext->tXbirContext.qSectorSize
                )
            || (
                    pLS->tVerifyFlashArgsIn.qNumPEBs
                    !=
                    pContext->tXbirContext.qNumSectors
                )
            || (
                    pLS->tVerifyFlashArgsIn.qSubsectorSize
                    <
                    sizeof(ReservedBlockEntry
                    )
                )
        ) {
        TRACE();
        return ERR_BOARD_FLASH;
    }

    TRACE();
    return ERR_NONE;
}

static uint32_t agentErasePEBs(
    ProgEnvFuncs* pPEF,
    uint32_t qPEBStart,
    uint32_t qPEBEnd,
    OUTPUT_PARAMETER volatile uint32_t* pqErrBSP
    )
{
    uint32_t qRet;
    PEFContext* pContext = pPEF->pContext;

    TRACE();
    *pqErrBSP = (uint32_t)FlashErase(
                    &pContext->tXbirContext,
                    qPEBStart * pContext->tXbirContext.qSectorSize,
                    (
                        (qPEBEnd - qPEBStart + 1)
                        *
                        pContext->tXbirContext.qSectorSize
                    )
                );

    qRet = ERR_NONE;
    return qRet;
}

static uint32_t agentReadPEB(
    ProgEnvFuncs* pPEF,
    uint32_t qPEBNumber,
    OUTPUT_PARAMETER volatile uint32_t* pqErrBSP,
    OUTPUT_PARAMETER uint8_t* pbDst
    )
{
    uint32_t qRet;
    PEFContext* pContext = pPEF->pContext;
    uint32_t qStart;

    TRACE();
    qStart = qPEBNumber * pContext->tXbirContext.qSectorSize;

    *pqErrBSP = (uint32_t)FlashRead(
                &pContext->tXbirContext,    /* XbirContext* pContext, */
                qStart,                     /* u32 Address, */
                pContext->tXbirContext.qSectorSize, /* u32 ByteCount, */
                pbDst                       /* u8 *ReadBfrPtr */
                );

    if((u32)XST_SUCCESS == *pqErrBSP) {
        qRet = ERR_NONE;
    } else {
        qRet = ERR_BSP;
    }
    return qRet;
}

static uint32_t agentWritePEB(
    ProgEnvFuncs* pPEF,
    uint32_t qPEBNumber,
    const uint8_t* pbSrc,
    OUTPUT_PARAMETER volatile uint32_t* pqErrBSP
    )
{
    uint32_t qRet;
    PEFContext* pContext = pPEF->pContext;
    uint32_t qNumPages;
    uint32_t qPage;
    uint32_t qStart;

    TRACE();
    qNumPages =
        pContext->tXbirContext.qSectorSize
        /
        pContext->tXbirContext.qPageSize
        ;
    qStart = qPEBNumber * pContext->tXbirContext.qSectorSize;

    for(qPage = 0; qPage < qNumPages; qPage++) {
        *pqErrBSP = (uint32_t)FlashWrite(
                    &pContext->tXbirContext,    /* PEFContext* pContext, */
                    qStart,                     /* u32 Address, */
                    pContext->tXbirContext.qPageSize,/* u32 ByteCount, */
                    (uint8_t*)pbSrc             /* u8 *WriteBfrPtr */
                    );
        if((u32)XST_SUCCESS != *pqErrBSP) {
            break;
        }
        qStart += pContext->tXbirContext.qPageSize;
        pbSrc += pContext->tXbirContext.qPageSize;
    }

    if((u32)XST_SUCCESS == *pqErrBSP) {
        qRet = ERR_NONE;
    } else {
        qRet = ERR_BSP;
    }
    return qRet;
}

static uint32_t agentGetMultibootBytes(
    ProgEnvFuncs* pPEF,
    OUTPUT_PARAMETER volatile uint32_t* pqErrBSP,
    OUTPUT_PARAMETER uint32_t* pqMultibootBytes
    )
{
    TRACE();
    *pqErrBSP = 0;
    *pqMultibootBytes = *(volatile uint32_t*)(XPS_PMC_GLOBAL_BASEADDRESS + 0x4) * 32 * 1024;

    return (uint32_t)ERR_NONE;
}

void agentInitPEFH(PEFContext* pContext)
{
    ProgEnvFuncs* pPEF;
    uint32_t q;

    TRACE();

    memset(pContext,0,sizeof(PEFContext));

#if defined(r5_0) || defined(r5_1)
    (void)Xil_SetMPURegion((INTPTR)0xfffc0000,256*1024,STRONG_ORDERD_SHARED|PRIV_RW_USER_RW);
    TRACE();
    (void)Xil_SetMPURegion((INTPTR)0xf000,4*1024,STRONG_ORDERD_SHARED|PRIV_RW_USER_RW);
    TRACE();
#endif

    pPEF = &pContext->tPEF;
    pPEF->pContext = pContext;
    pPEF->pMA = &g_tMessageArea;
    pPEF->pAA = &g_tAuxArea0;
    pPEF->pPA = &g_tAuxArea1;
    TRACE();

    pPEF->pMA->tCommandArea.qCommand = 0xCA11AB1E;
    for(q = 0; q < COMMAND_MAX_ARGS; q++) {
        pPEF->pMA->tCommandArea.pqArgs[q] = 0xCA11AB1E;
    }
    pPEF->pMA->tCommandArea.qChecksum = 0xCA11AB1E;
    TRACE();

    pPEF->pMA->tResponseArea.qErrCode = 0x5CA1AB1E;
    pPEF->pMA->tResponseArea.qErrBSP = 0x5CA1AB1E;
    for(q = 0; q < RESPONSE_MAX_DATA; q++) {
        pPEF->pMA->tResponseArea.pqData[q] = 0x5CA1AB1E;
    }
    pPEF->pMA->tResponseArea.qChecksum = 0x5CA1AB1E;
    pPEF->pMA->tResponseArea.qFEEDC0DE = 0x5CA1AB1E;
    TRACE();

    for(q = 0; q < MAX_PEB_SIZE_BYTES; NOTHING) {
        pPEF->pAA->pbData[q++] = 0x5E;
        pPEF->pAA->pbData[q++] = 0x00;
        pPEF->pAA->pbData[q++] = 0xAB;
        pPEF->pAA->pbData[q++] = 0x0C;
    }
    TRACE();
    for(q = 0; q < MAX_PEB_SIZE_BYTES; NOTHING) {
        pPEF->pPA->pbData[q++] = 0x11;
        pPEF->pPA->pbData[q++] = 0xBA;
        pPEF->pPA->pbData[q++] = 0xDD;
        pPEF->pPA->pbData[q++] = 0x00;
    }
    TRACE();

    pPEF->rDataBarrier = agentDataBarrier;
    pPEF->rVerifyFlash = agentVerifyFlash;
    pPEF->rErasePEBs = agentErasePEBs;
    pPEF->rReadPEB = agentReadPEB;
    pPEF->rWritePEB = agentWritePEB;
    pPEF->rGetMultibootBytes = agentGetMultibootBytes;

    TRACE();
    InitPEF(pPEF);
}
