/*
 * Copyright 2021-2022 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef PROTOCOL_H__
#define PROTOCOL_H__

#define MAX_PEB_SIZE_BYTES   (128*1024)

struct CommandArea;
typedef struct CommandArea CommandArea;

struct ResponseArea;
typedef struct ResponseArea ResponseArea;

struct MessageArea;
typedef struct MessageArea MessageArea;

struct AuxArea;
typedef struct AuxArea AuxArea;

typedef enum ERR {
    ERR_NONE = 0,
    ERR_SEQ,
    ERR_COMMAND,
    ERR_BOARD_FLASH,
    ERR_PEB_NUMBER,
    ERR_SUBSECTOR_NUMBER,
    ERR_ERASE,
    ERR_READ,
    ERR_WRITE,
    ERR_FLASH_UNVERIFIED,
    ERR_PEB_SIZE,
    ERR_SUBSECTOR_SIZE,
    ERR_IMAGE_SIZE,
    ERR_CHECKSUM,
    ERR_MULTIBOOT_BYTES,
    ERR_IMAGE_ALIGNMENT,
    ERR_BSP,
    ERR_TIMEOUT,
    ERR_INTERNAL = 0x7fffffff
} ERR;
#define ERR_STRING_BUFSIZE 32
extern const char* GetErrString(ERR eErr,OUTPUT_PARAMETER char pcErrString[ERR_STRING_BUFSIZE]);

typedef enum COMMAND {
    COMMAND_INVALID = 0,
    COMMAND_VERIFY_FLASH,
    COMMAND_VERIFY_IMAGE,
    COMMAND_GET_MULTIBOOT_BYTES,
    COMMAND_ERASE_PEBS,
    COMMAND_READ_PEB,
    COMMAND_WRITE_PEB,
    COMMAND_ERASE_SUBSECTOR,
    COMMAND_READ_SUBSECTOR,
    COMMAND_WRITE_SUBSECTOR,
    COMMAND_SIM_STOP    /* Simulation only */
} COMMAND;
#define COMMAND_STRING_BUFSIZE 32
extern const char* GetCommandString(COMMAND eCommand,OUTPUT_PARAMETER char pcCommandString[COMMAND_STRING_BUFSIZE]);

typedef enum CONNECTION_MODE {
    CONNECTION_MODE_LOWER,
    CONNECTION_MODE_DUAL_PARALLEL
} CONNECTION_MODE;

#define COMMAND_MAX_ARGS    ((512-(4*2))/4)
struct CommandArea {
    volatile uint32_t qCommand;
    volatile uint32_t pqArgs[COMMAND_MAX_ARGS];
    volatile uint32_t qChecksum;
};

#define FEEDC0DE       0xFEEDC0DE
#define RESPONSE_MAX_DATA   ((512-(4*4))/4)
struct ResponseArea {
    volatile uint32_t qErrCode;
    volatile uint32_t qErrBSP;
    volatile uint32_t pqData[RESPONSE_MAX_DATA];
    volatile uint32_t qChecksum;
    uint32_t qFEEDC0DE;
};

struct MessageArea {
    CommandArea tCommandArea;
    ResponseArea tResponseArea;
    uint32_t pqReserved[(3*1024)/4];
};

struct AuxArea {
    uint8_t pbData[MAX_PEB_SIZE_BYTES];
};

/*
 * Each R5 has a TCM. Each TCM is 128K, divided into A and B, each 64K.
 * R5 local address space:
 *  TCM A is 64K@0x0, while B is at 64K@0x20000.
 *  NOTE: The R5 local address space has a hole from 0x10000 to 0x20000.
 *
 * The R5s can be run in lock-step or split mode.
 * The global base address is 0xFFE0_0000 in both cases, but the locations
 * of {A,B}{0,1} vary based on mode, as shown below.
 *
 * TCM global address map for reference (TRM, table 56):
 * Split:
 *      A0      0   -- r5 #0
 *      B0      2   -- r5 #0    => Sync word is at FFE2F3FC
 *      A1      9   -- r5 #1
 *      B1      B   -- r5 #1
 *
 * Lockstep:
 *      A0      0   -- a72 #0
 *      A1      1   -- a72 #0   => Sync word is at FFE1F3FC
 *      B0      2   -- a72 #1
 *      B1      3   -- a72 #1
 *
 * The message area for communication with the PCIe host is 4K (0x1000) bytes.
 * Unfortunately, its global address (from PCIe, e.g.) varies based on whether
 * the chosen agent is R5#0 or A72#0. When the agent is R5#0, we use the split
 * mode, otherwise we use the locked mode
 *
 * Agent is R5#0: (split map)
 *     The area is placed at the end of B0 by the linker script
 *     Its local address is 0x2F000
 *     Its global address is 0xFFE2F000
 * Agent is A72#0: (locked map)
 *     A0 and A1 are used by A72#0
 *     The area is placed at the end of A1 by the linker script
 *     Its global address is 0xFFE1F000
 *
 * The sync word is at on offset 0x3fc within that message area.
 * Agent r5#0: this is 0x2F3FC in its local address space, 0xFFE2F3FC global.
 * Agent a72#0: This is global address 0xFFE1F3FC.
 *
 * The x86 host does not know a-priori which agent is being used. Thus
 * it has to check both locations for the sync code in order to work
 */

/*
 * The offsets are from the TCM base, 0xFFE0_0000
 * The TCM offset in the PCIe BAR is specified in the board configuration
 */
#define AGENT_MESSAGE_AREA_OFFSET_A72   0x1F000
#define AGENT_MESSAGE_AREA_OFFSET_R5    0x2F000
#define AGENT_AUX_AREA_OFFSET           (0xfffc0000 - 0xffe00000)
#endif  /* PROTOCOL_H__ */
